from hypothesis import given
from hypothesis import strategies as st
from sqlalchemy import select

from nextline_rdb.db import DB
from nextline_rdb.utils.strategies import st_none_or

from ... import Model, Script
from .. import st_model_script


@given(st.data())
async def test_options(data: st.DataObject) -> None:
    current = data.draw(st_none_or(st.booleans()))
    script = data.draw(st_model_script(current=current))
    if current is not None:
        assert script.current == current
    compile(script.script, '<string>', 'exec')


@given(script=st_model_script())
async def test_db(script: Script) -> None:
    async with DB(use_migration=False, model_base_class=Model) as db:
        async with db.session.begin() as session:
            session.add(script)
        async with db.session() as session:
            select_script = select(Script)
            script_ = await session.scalar(select_script)

    assert repr(script) == repr(script_)



"""migrate data

Revision ID: 68cb1c590d1f
Revises: 245c4152e8d7
Create Date: 2024-02-03 14:13:13.174733

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.orm.session import Session

from nextline_rdb.alembic.models.rev_245c4152e8d7 import CurrentScript, Script

# revision identifiers, used by Alembic.
revision = '68cb1c590d1f'
down_revision = '245c4152e8d7'
branch_labels = None
depends_on = None


def upgrade():
    with Session(bind=op.get_bind()) as session, session.begin():
        select_script = sa.select(Script).where(Script.current.is_(True))
        script = session.execute(select_script).scalar_one_or_none()
        if script is not None:
            current_script = CurrentScript(script_id=script.id)
            script.current = False
            session.add(current_script)


def downgrade():
    with Session(bind=op.get_bind()) as session, session.begin():
        select_current_script = sa.select(CurrentScript)
        current_script = session.execute(select_current_script).scalar_one_or_none()
        if current_script is not None:
            current_script.script.current = True
            session.delete(current_script)



from collections.abc import Sized

from hypothesis import given
from hypothesis import strategies as st

from nextline_rdb.pagination import Sort, SortField
from nextline_rdb.utils.strategies import st_graphql_ints, st_none_or

from .models import Entity


def st_entity() -> st.SearchStrategy[Entity]:
    return st.builds(
        Entity,
        num=st_none_or(st_graphql_ints()),
        txt=st_none_or(st.text()),
    )


@st.composite
def st_sort(draw: st.DrawFn) -> Sort:
    FIELDS = ('id', 'num', 'txt')
    fields = draw(st.lists(st.sampled_from(FIELDS), max_size=len(FIELDS), unique=True))
    return [SortField(field, draw(st.booleans())) for field in fields]


@given(st_sort())
def test_st_sort(sort: Sort):
    # ic(sort)
    pass


def cmp(e1: Entity, e2: Entity, sort: Sort | None) -> int:
    '''A comparison function to be used as `comp_to_key(partial(cmp, sort=sort))`'''
    for f in sort or []:
        a = getattr(e1, f.field)
        b = getattr(e2, f.field)
        match a, b:
            case None, None:
                continue
            case None, _:
                return 1 if f.desc else -1
            case _, None:
                return -1 if f.desc else 1
            case a, b if a < b:
                return 1 if f.desc else -1
            case a, b if a > b:
                return -1 if f.desc else 1
            case _:
                pass
    return 0


def st_idx(items: Sized) -> st.SearchStrategy[int | None]:
    '''A strategy that generates an index of `items` or `None`'''
    return st_none_or(st.sampled_from(range(len(items)))) if items else st.none()


def st_length(max_value: int, allow_none: bool = True) -> st.SearchStrategy[int | None]:
    '''A strategy that generates a natural number <= `max_value` or `None`'''
    st_int = st.integers(min_value=0, max_value=max_value)
    return st_none_or(st_int) if allow_none else st_int



from .connection import Connection, Edge, PageInfo, query_connection
from .db import load_connection

__all__ = [
    "Edge",
    "PageInfo",
    "Connection",
    "query_connection",
    "load_connection",
]



import datetime

from hypothesis import given
from hypothesis import strategies as st
from sqlalchemy import select
from sqlalchemy.orm import selectinload

from nextline_rdb.db import DB

from .. import Model, Prompt
from ..strategies import st_model_prompt


@given(st.data())
async def test_repr(data: st.DataObject):
    async with DB(use_migration=False, model_base_class=Model) as db:
        async with db.session.begin() as session:
            model = data.draw(st_model_prompt())
            session.add(model)

        async with db.session() as session:
            rows = await session.scalars(
                select(Prompt).options(
                    selectinload(Prompt.run), selectinload(Prompt.trace)
                )
            )
            for row in rows:
                repr_ = repr(row)
                assert Prompt, datetime  # type: ignore[truthy-function]
                assert repr_ == repr(eval(repr_))



import asyncio
import logging
import logging.config

from alembic import context
from sqlalchemy import Connection
from sqlalchemy.ext.asyncio import create_async_engine

from nextline_rdb import models

config = context.config
url = config.get_main_option("sqlalchemy.url")

if config.config_file_name:
    # from logging_tree import printout
    # printout()
    if config.cmd_opts is None:
        # Presumably, alembic.command.upgrade() or other functions from
        # alembic.command are called rather than the alembic command is
        # executed from a shell. Do not configure logging here so as to avoid
        # overriding the logging config.
        pass
    else:
        logging.config.fileConfig(config.config_file_name)


logger = logging.getLogger(__name__)
logger.info(f'DB URL: {url}')


target_metadata = models.Model.metadata


def run_migrations_offline() -> None:
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    context.configure(
        url=url,
        target_metadata=target_metadata,
        literal_binds=True,
        dialect_opts={"paramstyle": "named"},
    )

    with context.begin_transaction():
        context.run_migrations()


def do_run_migrations(connection: Connection) -> None:
    context.configure(
        connection=connection,
        target_metadata=target_metadata,
        render_as_batch=True,
    )

    with context.begin_transaction():
        context.run_migrations()


async def run_async_migrations() -> None:
    assert url is not None
    connectable = create_async_engine(url)

    async with connectable.connect() as connection:
        await connection.run_sync(do_run_migrations)

    await connectable.dispose()


def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    asyncio.run(run_async_migrations())


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



from hypothesis import given
from hypothesis import strategies as st

from nextline_rdb.utils import mark_last


@given(st.lists(st.integers()))
def test_one(items: list[int]):
    it = iter(items)
    actual = list(mark_last(it))

    expected = [(i == len(items) - 1, e) for i, e in enumerate(items)]

    assert actual == expected



import asyncio
from datetime import timezone

from nextline.plugin.spec import Context, hookimpl
from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import selectinload

from nextline_rdb.db import DB
from nextline_rdb.models import CurrentScript, Run, Script


class WriteRunTable:
    def __init__(self, db: DB) -> None:
        self._db = db

    @hookimpl
    async def on_initialize_run(self, context: Context) -> None:
        assert (run_arg := context.run_arg)
        run_no = run_arg.run_no
        if isinstance(run_arg.statement, str):
            statement = run_arg.statement
            await self._on_initialize_run_with_statement(run_no, statement)
        else:
            await self._on_initialize_run_without_statement(run_no)

    async def _on_initialize_run_without_statement(self, run_no: int) -> None:
        async with self._db.session.begin() as session:
            current_script = await self._load_current_script(session)
            if current_script is not None:
                await session.delete(current_script)
            run = Run(run_no=run_no, state='initialized')
            session.add(run)

    async def _on_initialize_run_with_statement(
        self, run_no: int, statement: str
    ) -> None:
        async with self._db.session.begin() as session:
            current_script = await self._load_current_script(session)
            if current_script is not None:
                if current_script.script.script == statement:
                    script = current_script.script
                else:
                    script = Script(script=statement)
                    current_script.script = script
            else:
                script = Script(script=statement)
                current_script = CurrentScript(script=script)
                session.add(current_script)
            run = Run(run_no=run_no, state='initialized', script=script)
            session.add(run)

    async def _load_current_script(self, session: AsyncSession) -> CurrentScript | None:
        stmt = select(CurrentScript).options(selectinload(CurrentScript.script))
        return (await session.execute(stmt)).scalar_one_or_none()

    @hookimpl
    async def on_start_run(self, context: Context) -> None:
        assert (run_arg := context.run_arg)
        assert (running_process := context.running_process)
        started_at = running_process.process_created_at
        assert started_at.tzinfo is timezone.utc
        started_at = started_at.replace(tzinfo=None)
        run_no = run_arg.run_no
        async with self._db.session.begin() as session:
            stmt = select(Run).filter_by(run_no=run_no)
            while not (run := (await session.execute(stmt)).scalar_one_or_none()):
                await asyncio.sleep(0)
            run.state = 'running'
            run.started_at = started_at

    @hookimpl
    async def on_end_run(self, context: Context) -> None:
        assert (run_arg := context.run_arg)
        assert (exited_process := context.exited_process)
        assert (returned := exited_process.returned)
        ended_at = exited_process.process_exited_at
        assert ended_at.tzinfo is timezone.utc
        ended_at = ended_at.replace(tzinfo=None)
        run_no = run_arg.run_no
        async with self._db.session.begin() as session:
            stmt = select(Run).filter_by(run_no=run_no)
            while not (run := (await session.execute(stmt)).scalar_one_or_none()):
                await asyncio.sleep(0)
            run.state = 'finished'
            run.ended_at = ended_at
            run.exception = returned.fmt_exc



import asyncio

from nextline.events import OnEndPrompt, OnStartPrompt
from nextline.plugin.spec import hookimpl
from sqlalchemy import select

from nextline_rdb.db import DB
from nextline_rdb.models import Prompt, Run, Trace


class WritePromptTable:
    def __init__(self, db: DB) -> None:
        self._db = db

    @hookimpl
    async def on_start_prompt(self, event: OnStartPrompt) -> None:
        async with self._db.session.begin() as session:
            select_runs = select(Run).filter_by(run_no=event.run_no)
            while not (
                run := (await session.execute(select_runs)).scalar_one_or_none()
            ):
                await asyncio.sleep(0)
            select_traces = select(Trace).filter_by(
                run_no=event.run_no, trace_no=event.trace_no
            )
            while not (
                trace := (await session.execute(select_traces)).scalar_one_or_none()
            ):
                await asyncio.sleep(0)
            prompt = Prompt(
                run_no=event.run_no,
                trace_no=event.trace_no,
                prompt_no=event.prompt_no,
                open=True,
                event=event.event,
                file_name=event.file_name,
                line_no=event.line_no,
                stdout=event.prompt_text,
                started_at=event.started_at,
                run=run,
                trace=trace,
            )
            session.add(prompt)

    @hookimpl
    async def on_end_prompt(self, event: OnEndPrompt) -> None:
        async with self._db.session.begin() as session:
            stmt = select(Prompt).filter_by(
                run_no=event.run_no, prompt_no=event.prompt_no
            )
            while not (prompt := (await session.execute(stmt)).scalar_one_or_none()):
                await asyncio.sleep(0)
            prompt.open = False
            prompt.command = event.command
            prompt.ended_at = event.ended_at



import base64
from typing import TypedDict

import pytest
import strawberry
from hypothesis import given, note
from hypothesis import strategies as st

from nextline_rdb.db import DB
from nextline_rdb.models import Run
from nextline_rdb.models.strategies import st_model_run_list
from nextline_rdb.schema import Query

from ..graphql import QUERY_RDB_RUNS


def Cursor(i: int) -> str:
    return base64.b64encode(f'{i}'.encode()).decode()


class Variables(TypedDict, total=False):
    before: str | None
    after: str | None
    first: int | None
    last: int | None


class PageInfo(TypedDict):
    hasPreviousPage: bool
    hasNextPage: bool
    startCursor: str | None
    endCursor: str | None


class Node(TypedDict):
    id: int
    runNo: int


class Edge(TypedDict):
    cursor: str
    node: Node


@given(runs=st_model_run_list(generate_traces=False, min_size=0, max_size=12))
async def test_all(runs: list[Run]):
    schema = strawberry.Schema(query=Query)

    async with DB() as db:
        async with db.session.begin() as session:
            session.add_all(runs)
        note(f'runs: {runs}')

        nodes_saved = [Node(id=run.id, runNo=run.run_no) for run in runs]
        note(f'nodes_saved: {nodes_saved}')

        resp = await schema.execute(QUERY_RDB_RUNS, context_value={'db': db})
        assert resp.data

        all_runs = resp.data['rdb']['runs']
        page_info: PageInfo = all_runs['pageInfo']
        edges: list[Edge] = all_runs['edges']

        if edges:
            assert page_info['startCursor'] == edges[0]['cursor']
            assert page_info['endCursor'] == edges[-1]['cursor']

        nodes = [edge['node'] for edge in edges]

        assert nodes == nodes_saved


@given(
    runs=st_model_run_list(generate_traces=False, min_size=0, max_size=12),
    first=st.integers(min_value=1, max_value=15),
)
async def test_forward(runs: list[Run], first: int):
    schema = strawberry.Schema(query=Query)

    async with DB() as db:
        async with db.session.begin() as session:
            session.add_all(runs)
        note(f'runs: {runs}')

        nodes_saved = [Node(id=run.id, runNo=run.run_no) for run in runs]
        note(f'nodes_saved: {nodes_saved}')

        after = None
        has_next_page = True
        nodes = list[Node]()
        while has_next_page:
            variables = Variables(after=after, first=first)

            resp = await schema.execute(
                QUERY_RDB_RUNS,
                variable_values=dict(variables),
                context_value={'db': db},
            )
            assert resp.data

            all_runs = resp.data['rdb']['runs']
            page_info: PageInfo = all_runs['pageInfo']
            edges: list[Edge] = all_runs['edges']

            has_next_page = page_info['hasNextPage']
            after = page_info['endCursor']

            if not nodes_saved:
                assert not edges
                assert not has_next_page
                assert after is None

            if edges:
                assert after == edges[-1]['cursor']

            nodes.extend(edge['node'] for edge in edges)

        assert nodes == nodes_saved


@given(
    runs=st_model_run_list(generate_traces=False, min_size=0, max_size=12),
    last=st.integers(min_value=1, max_value=15),
)
async def test_backward(runs: list[Run], last: int):
    schema = strawberry.Schema(query=Query)

    async with DB() as db:
        async with db.session.begin() as session:
            session.add_all(runs)
        note(f'runs: {runs}')

        nodes_saved = [Node(id=run.id, runNo=run.run_no) for run in runs]
        note(f'nodes_saved: {nodes_saved}')

        before = None
        has_previous_page = True
        nodes = list[Node]()
        while has_previous_page:
            variables = Variables(before=before, last=last)

            resp = await schema.execute(
                QUERY_RDB_RUNS,
                variable_values=dict(variables),
                context_value={'db': db},
            )
            assert resp.data

            all_runs = resp.data['rdb']['runs']
            page_info: PageInfo = all_runs['pageInfo']
            edges: list[Edge] = all_runs['edges']

            has_previous_page = page_info['hasPreviousPage']
            before = page_info['startCursor']

            if not nodes_saved:
                assert not edges
                assert not has_previous_page
                assert before is None

            if edges:
                assert before == edges[0]['cursor']

            nodes.extend(edge['node'] for edge in reversed(edges))

        assert nodes == list(reversed(nodes_saved))


@given(runs=st_model_run_list(generate_traces=False, min_size=0, max_size=12))
async def test_cursor(runs: list[Run]):
    schema = strawberry.Schema(query=Query)

    async with DB() as db:
        async with db.session.begin() as session:
            session.add_all(runs)
        note(f'runs: {runs}')

        nodes_saved = [Node(id=run.id, runNo=run.run_no) for run in runs]
        note(f'nodes_saved: {nodes_saved}')

        resp = await schema.execute(QUERY_RDB_RUNS, context_value={'db': db})
        assert resp.data

        all_runs = resp.data['rdb']['runs']
        edges: list[Edge] = all_runs['edges']

        nodes = [edge['node'] for edge in edges]
        cursors = [edge['cursor'] for edge in edges]

        assert cursors == [Cursor(node['id']) for node in nodes]


params = [
    pytest.param(
        Variables(first=5, last=5),
        id='first-and-last',
    ),
    pytest.param(
        Variables(before=Cursor(31), first=5),
        id='before-and-first',
    ),
    pytest.param(
        Variables(after=Cursor(20), last=5),
        id='after-and-last',
    ),
    pytest.param(
        Variables(before=Cursor(31), after=Cursor(20)),
        id='before-and-after',
    ),
    pytest.param(
        Variables(before=Cursor(31), after=Cursor(20), first=5, last=5),
        id='all',
    ),
]


@pytest.mark.parametrize('variables', params)
async def test_error(variables: Variables) -> None:
    schema = strawberry.Schema(query=Query)

    async with DB() as db:
        resp = await schema.execute(
            QUERY_RDB_RUNS,
            variable_values=dict(variables),
            context_value={'db': db},
        )

        assert not resp.data
        assert resp.errors



import pytest
from hypothesis import given
from sqlalchemy.exc import IntegrityError

from nextline_rdb.db import DB
from nextline_rdb.utils import class_name_and_primary_keys_of, load_all

from .. import CurrentScript, Model, Script
from ..strategies import st_model_script


@given(script=st_model_script())
async def test_repr(script: Script) -> None:
    async with DB(use_migration=False, model_base_class=Model) as db:
        async with db.session.begin() as session:
            session.add(script)
            added = list(session.new)
        added = sorted(added, key=class_name_and_primary_keys_of)
        repr_added = repr(added)

        async with db.session() as session:
            loaded = await load_all(session, Model)
            repr_loaded = repr(loaded)

            assert repr_added == repr_loaded
            assert CurrentScript  # type: ignore[truthy-function]
            assert repr_loaded == repr(eval(repr_loaded))


async def test_at_most_one_current_script() -> None:
    async with DB(use_migration=False, model_base_class=Model) as db:
        async with db.session() as session:
            current_script = CurrentScript(script=Script(script=''))
            session.add(current_script)
            await session.commit()

            current_script = CurrentScript(script=Script(script=''))
            session.add(current_script)
            with pytest.raises(IntegrityError):
                await session.commit()



import base64
from collections.abc import Callable
from functools import partial
from typing import Optional, Type, TypeVar

from nextline_rdb import models as db_models
from nextline_rdb.db import DB
from nextline_rdb.pagination import Sort, load_models

from .connection import Connection, Edge, query_connection

_T = TypeVar('_T')


def encode_id(id: int) -> str:
    return base64.b64encode(f'{id}'.encode()).decode()


def decode_id(cursor: str) -> int:
    return int(base64.b64decode(cursor).decode())


async def load_connection(
    db: DB,
    Model: Type[db_models.Model],
    id_field: str,
    create_node_from_model: Callable[..., _T],
    *,
    sort: Optional[Sort] = None,
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> Connection[_T]:
    query_edges = partial(
        load_edges,
        db=db,
        Model=Model,
        id_field=id_field,
        create_node_from_model=create_node_from_model,
        sort=sort,
    )

    return await query_connection(
        query_edges,
        before,
        after,
        first,
        last,
    )


async def load_edges(
    db: DB,
    Model: Type[db_models.Model],
    id_field: str,
    create_node_from_model: Callable[..., _T],
    *,
    sort: Optional[Sort] = None,
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> list[Edge[_T]]:
    async with db.session() as session:
        models = await load_models(
            session,
            Model,
            id_field,
            sort=sort,
            before=before if before is None else decode_id(before),
            after=after if after is None else decode_id(after),
            first=first,
            last=last,
        )

    nodes = [create_node_from_model(m) for m in models]

    edges = [Edge(node=n, cursor=encode_id(getattr(n, id_field))) for n in nodes]

    return edges



import datetime as dt
from collections.abc import Iterable
from typing import Callable, Optional, cast

from hypothesis import strategies as st

from nextline_rdb.utils import mark_last
from nextline_rdb.utils.strategies import st_datetimes, st_graphql_ints, st_none_or

from .. import Run, Script
from .utils import st_started_at_ended_at


@st.composite
def st_model_run(
    draw: st.DrawFn,
    run_no: Optional[int] = None,
    min_run_no: Optional[int] = None,
    max_run_no: Optional[int] = None,
    min_started_at: Optional[dt.datetime] = None,
    max_started_at: Optional[dt.datetime] = None,
    min_ended_at: Optional[dt.datetime] = None,
    max_ended_at: Optional[dt.datetime] = None,
    script: Optional[Script] = None,
    generate_traces: bool = True,
) -> Run:
    from .st_prompt import st_model_prompt_list
    from .st_script import st_model_script
    from .st_stdout import st_model_stdout_list
    from .st_trace import st_model_trace_list

    def st_run_no() -> st.SearchStrategy[int]:
        if run_no is not None:
            return st.just(run_no)
        min_ = min_run_no if min_run_no is not None else 1
        return st_graphql_ints(min_value=min_, max_value=max_run_no)

    run_no = draw(st_run_no())

    state = draw(st_none_or(st.text()))

    started_at, ended_at = draw(
        st_started_at_ended_at(
            min_start=min_started_at,
            max_start=max_started_at,
            min_end=min_ended_at,
            max_end=max_ended_at,
        )
    )

    if script is None:
        script = draw(st_none_or(st_model_script()))

    exception = draw(st_none_or(st.text()))

    run = Run(
        run_no=run_no,
        state=state,
        started_at=started_at,
        ended_at=ended_at,
        script=script,
        exception=exception,
    )

    if generate_traces:
        draw(st_model_trace_list(run=run, min_size=1, max_size=4))
        draw(st_model_prompt_list(run=run, min_size=0, max_size=8))
        draw(st_model_stdout_list(run=run, min_size=0, max_size=5))

    return run


@st.composite
def st_model_run_list(
    draw: st.DrawFn,
    generate_traces: bool = False,
    min_size: int = 0,
    max_size: Optional[int] = None,
) -> list[Run]:
    from .st_script import st_model_script

    run_nos = draw(
        st.lists(
            st_graphql_ints(min_value=1),
            min_size=min_size,
            max_size=max_size,
            unique=True,
        ).map(cast(Callable[[Iterable[int]], list[int]], sorted))
    )

    runs = list[Run]()
    scripts = list[Script]()
    min_started_at = None
    for last, run_no in mark_last(run_nos):
        min_started_at = min_started_at or draw(st_datetimes())

        if scripts:
            script = draw(
                st.one_of(
                    st_none_or(st_model_script(current=False)),
                    st.sampled_from(scripts),
                )
            )
        else:
            script = draw(st_none_or(st_model_script(current=False)))
            if script is not None:
                scripts.append(script)

        run = draw(
            st_model_run(
                run_no=run_no,
                script=script,
                generate_traces=generate_traces,
                min_started_at=min_started_at,
            )
        )
        assert run.run_no == run_no

        if run.script is not None:
            run.script.current = last

        if run.started_at is not None:
            min_started_at = run.started_at + dt.timedelta(seconds=1)
        # if run.ended_at is not None:
        #     min_started_at = run.ended_at + dt.timedelta(seconds=1)
        runs.append(run)
    return runs



from hypothesis import given
from nextline import Nextline

from nextline_rdb.db import DB
from nextline_rdb.init import initialize_nextline
from nextline_rdb.models import Run
from nextline_rdb.models.strategies import st_model_run_list
from nextline_rdb.utils.strategies import st_python_scripts


@given(
    runs=st_model_run_list(generate_traces=False, min_size=0, max_size=2),
    default_statement=st_python_scripts(),
)
async def test_initialize_nextline(runs: list[Run], default_statement: str) -> None:
    last_run = runs[-1] if runs else None

    expected_run_no = last_run.run_no + 1 if last_run else 1

    current_scripts = [r.script.script for r in runs if r.script and r.script.current]
    expected_statement = (
        default_statement if not current_scripts else current_scripts[-1]
    )

    nextline = Nextline(statement=default_statement)

    async with DB() as db:
        async with db.session.begin() as session:
            session.add_all(runs)
        await initialize_nextline(nextline, db)

    init_options = nextline._init_options

    assert init_options.run_no_start_from == expected_run_no
    assert init_options.statement == expected_statement



from collections.abc import AsyncIterator, Mapping, MutableMapping
from pathlib import Path
from typing import Optional

from apluggy import asynccontextmanager
from dynaconf import Dynaconf, Validator
from nextlinegraphql.hook import spec

from . import write
from .db import DB
from .init import initialize_nextline
from .schema import Mutation, Query, Subscription

HERE = Path(__file__).resolve().parent
DEFAULT_CONFIG_PATH = HERE / 'default.toml'

assert DEFAULT_CONFIG_PATH.is_file()

PRELOAD = (str(DEFAULT_CONFIG_PATH),)
SETTINGS = ()
VALIDATORS = (Validator("DB.URL", must_exist=True, is_type_of=str),)


class Plugin:
    @spec.hookimpl
    def dynaconf_preload(self) -> Optional[tuple[str, ...]]:
        return PRELOAD

    @spec.hookimpl
    def dynaconf_settings_files(self) -> Optional[tuple[str, ...]]:
        return SETTINGS

    @spec.hookimpl
    def dynaconf_validators(self) -> Optional[tuple[Validator, ...]]:
        return VALIDATORS

    @spec.hookimpl
    def configure(self, settings: Dynaconf) -> None:
        self.url = settings.db['url']

    @spec.hookimpl
    def schema(self) -> tuple[type, type | None, type | None]:
        return (Query, Mutation, Subscription)

    @spec.hookimpl
    @asynccontextmanager
    async def lifespan(self, context: Mapping) -> AsyncIterator[None]:
        nextline = context['nextline']
        async with DB(self.url) as db:
            self._db = db
            await initialize_nextline(nextline, db)
            write.register(nextline=nextline, db=self._db)
            yield

    @spec.hookimpl
    def update_strawberry_context(self, context: MutableMapping) -> None:
        context['db'] = self._db



from alembic import command
from alembic.config import Config


def test_migration_up_to_date(alembic_config_temp_sqlite: Config) -> None:
    '''Run alembic check

    This test fails if the head of the migration versions is not consistent of the ORM
    models defined in src/nextline_rdb/models.

    - https://github.com/sqlalchemy/alembic/releases/tag/rel_1_9_0
    - https://github.com/sqlalchemy/alembic/issues/724
    - https://github.com/sqlalchemy/alembic/pull/1101
    '''
    config = alembic_config_temp_sqlite
    command.upgrade(config, "head")  # This line might be unnecessary
    command.check(config)



import pytest
from sqlalchemy.exc import IntegrityError

from nextline_rdb.db import DB

from .models import Bar, Model


async def test_foreign_key_constraints() -> None:
    '''Assert that foreign key constraints are enabled.

    The foreign key constraints are not enabled by default in SQLite.

    - https://www.sqlite.org/foreignkeys.html
    - https://docs.sqlalchemy.org/en/20/dialects/sqlite.html#foreign-key-support

    This test tries to insert a row with a foreign key that does not exist and confirms
    that an IntegrityError is raised.

    This test is copied from c5backup:
    https://github.com/simonsobs/c5backup/blob/v0.3.0/tests/test_db.py#L48
    '''
    async with DB(model_base_class=Model, use_migration=False) as db:
        async with db.session() as session:
            bar = Bar(foo_id=1)
            session.add(bar)

            with pytest.raises(IntegrityError, match=r'(?i)foreign.*key.*constraint'):
                await session.commit()



from logging import getLogger
from typing import Optional

from nextline import Nextline
from sqlalchemy import func, select
from sqlalchemy.orm import selectinload

from .db import DB
from .models import CurrentScript, Run


async def initialize_nextline(nextline: Nextline, db: DB) -> None:
    run_no = await _last_run_no(db)
    if run_no is not None:
        run_no += 1
        if run_no >= nextline._init_options.run_no_start_from:
            nextline._init_options.run_no_start_from = run_no
    script = await _current_script(db)
    if script is not None:
        nextline._init_options.statement = script


async def _last_run_no(db: DB) -> Optional[int]:
    async with db.session() as session:
        stmt = select(Run, func.max(Run.run_no))
        if run := (await session.execute(stmt)).scalar_one_or_none():
            return run.run_no
        logger = getLogger(__name__)
        msg = 'No previous runs were found in the DB'
        logger.info(msg)
        return None


async def _current_script(db: DB) -> Optional[str]:
    async with db.session() as session:
        stmt = select(CurrentScript).options(selectinload(CurrentScript.script))
        if current_script := (await session.execute(stmt)).scalar_one_or_none():
            return current_script.script.script
        logger = getLogger(__name__)
        msg = 'No current scripts were found in the DB'
        logger.info(msg)
        return None



from collections.abc import Callable
from pathlib import Path

import pytest
import tomli
import tomli_w
from async_asgi_testclient import TestClient
from nextlinegraphql import create_app
from nextlinegraphql.plugins.ctrl.graphql import MUTATE_RUN_AND_CONTINUE
from nextlinegraphql.plugins.graphql.test import gql_request

from nextline_rdb.db import DB
from nextline_rdb.models.strategies import st_model_run_list

from .schema.graphql import QUERY_RDB_CONNECTIONS


async def test_plugin(set_new_url: Callable[[], str]):
    # Enter some runs into the database.
    runs = st_model_run_list(generate_traces=False, max_size=2).example()

    url = set_new_url()
    async with DB(url) as db:
        async with db.session.begin() as session:
            session.add_all(runs)

    #
    expected_n_runs = len(runs) + 1  # +1 for the new run

    # Execute GraphQL queries.
    app = create_app()  # the plugin is loaded here
    async with TestClient(app) as client:
        data = await gql_request(client, MUTATE_RUN_AND_CONTINUE)
        assert data['runAndContinue']
        data = await gql_request(client, QUERY_RDB_CONNECTIONS)
        n_runs = len(data['rdb']['runs']['edges'])

    assert n_runs == expected_n_runs


def test_fixture(settings_path: Path, set_new_url: Callable[[], str]):
    from nextline_rdb import plugin

    assert str(settings_path) in plugin.SETTINGS

    settings = tomli.loads(settings_path.read_text())
    url_org = settings['db']['url']

    url = set_new_url()
    assert url != url_org

    settings = tomli.loads(settings_path.read_text())
    assert url == settings['db']['url']


@pytest.fixture(scope='module')
def set_new_url(
    settings_path: Path, tmp_url_factory: Callable[[], str]
) -> Callable[[], str]:
    def _f() -> str:
        url = tmp_url_factory()
        settings = tomli.loads(settings_path.read_text())
        settings['db']['url'] = url
        settings_path.write_text(tomli_w.dumps(settings))
        return url

    return _f


@pytest.fixture(autouse=True)
def monkeypatch_settings_path(monkeypatch: pytest.MonkeyPatch, settings_path: Path):
    from nextline_rdb import plugin

    monkeypatch.setattr(plugin, 'SETTINGS', (str(settings_path),))


@pytest.fixture(scope='module')
def settings_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    dir = tmp_path_factory.mktemp('config')
    settings_path = dir / 'settings.toml'
    url = 'sqlite+aiosqlite://'
    data = {'db': {'url': url}}
    toml_str = tomli_w.dumps(data)
    with settings_path.open('w') as f:
        f.write(toml_str)
    return settings_path


@pytest.fixture(scope='module')
def tmp_url_factory(tmp_path_factory: pytest.TempPathFactory) -> Callable[[], str]:
    def factory() -> str:
        dir = tmp_path_factory.mktemp('db')
        url = f'sqlite+aiosqlite:///{dir}/db.sqlite'
        return url

    return factory



from pathlib import Path

pwd = Path(__file__).resolve().parent

sub = pwd / 'queries'
QUERY_RDB_CONNECTIONS = (sub / 'RDBConnections.gql').read_text()
QUERY_RDB_RUNS = (sub / 'RDBRuns.gql').read_text()
QUERY_RDB_RUN = (sub / 'RDBRun.gql').read_text()

sub = pwd / 'mutations'
MUTATE_RDB_DELETE_RUNS = (sub / 'RDBDeleteRuns.gql').read_text()


# sub = pwd / 'subscriptions'



import datetime

from hypothesis import given
from sqlalchemy import select

from nextline_rdb.db import DB

from .. import Model, Run
from ..strategies import st_model_run


@given(run=st_model_run(generate_traces=False))
async def test_repr(run: Run):
    async with DB(use_migration=False, model_base_class=Model) as db:
        async with db.session.begin() as session:
            session.add(run)

        async with db.session() as session:
            rows = await session.scalars(select(Run))
            for row in rows:
                repr_ = repr(row)
                assert Run, datetime  # type: ignore[truthy-function]
                assert repr_ == repr(eval(repr_))


@given(run=st_model_run(generate_traces=True))
async def test_cascade(run: Run):
    async with DB(use_migration=False, model_base_class=Model) as db:
        async with db.session.begin() as session:
            session.add(run)

        async with db.session.begin() as session:
            select_run = select(Run)
            run = (await session.execute(select_run)).scalar_one()
            await session.delete(run)

        async with db.session() as session:
            select_traces = select(Run)
            traces = (await session.execute(select_traces)).scalars().all()
            assert not traces



from black import Mode, format_str
from hypothesis import strategies as st


@st.composite
def st_python_scripts(draw: st.DrawFn) -> str:
    types = ('print', 'sleep')
    lines = ['import time']
    n_lines = draw(st.integers(min_value=1, max_value=10))
    for _ in range(n_lines):
        type_ = draw(st.sampled_from(types))
        match type_:
            case 'print':
                text = draw(st.text(max_size=120))
                lines.append(f'print({text!r})')
            case 'sleep':
                secs = draw(st.integers(min_value=1, max_value=100)) / 10_000
                lines.append(f'time.sleep({secs:.5f})')

    code = '\n'.join(lines)
    return format_str(code, mode=Mode())



"""remove the field "current" from the table "script"

Revision ID: 4dc6a93dfed8
Revises: 68cb1c590d1f
Create Date: 2024-02-04 09:01:35.420772

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '4dc6a93dfed8'
down_revision = '68cb1c590d1f'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute('PRAGMA foreign_keys=OFF;')
    with op.batch_alter_table('script', schema=None) as batch_op:
        batch_op.drop_column('current')
    op.execute('PRAGMA foreign_keys=ON;')

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute('PRAGMA foreign_keys=OFF;')

    # Step 1: Add the column as nullable
    with op.batch_alter_table('script', schema=None) as batch_op:
        batch_op.add_column(sa.Column('current', sa.BOOLEAN(), nullable=True))

    # Step 2: Update existing rows to set the default value for the new column
    op.execute(
        '''
        UPDATE script
        SET current = False
    '''
    )

    # Step 3: Alter the column to be non-nullable
    with op.batch_alter_table('script', schema=None) as batch_op:
        batch_op.alter_column('current', nullable=False)

    # An alternative way that sets the default value in the database level
    # with op.batch_alter_table('script', schema=None) as batch_op:
    #     batch_op.add_column(
    #         sa.Column(
    #             'current',
    #             sa.BOOLEAN(),
    #             nullable=False,
    #             server_default=sa.sql.expression.false(),
    #         )
    #     )
    op.execute('PRAGMA foreign_keys=ON;')

    # ### end Alembic commands ###



from sqlalchemy.orm import Mapped, mapped_column

from .base import Model


class Hello(Model):
    __tablename__ = "hello"
    id: Mapped[int] = mapped_column(primary_key=True, index=True)
    message: Mapped[str | None]



__all__ = [
    'st_datetimes',
    'SQLITE_INT_MAX',
    'SQLITE_INT_MIN',
    'StMinMaxValuesFactory',
    'st_graphql_ints',
    'st_none_or',
    'st_ranges',
    'st_sqlite_ints',
    'st_python_scripts',
]


from .datetime import st_datetimes
from .misc import (
    SQLITE_INT_MAX,
    SQLITE_INT_MIN,
    StMinMaxValuesFactory,
    st_graphql_ints,
    st_none_or,
    st_ranges,
    st_sqlite_ints,
)
from .script import st_python_scripts



from hypothesis import given
from hypothesis import strategies as st

from nextline_rdb.utils import safe_max, safe_min


@given(st.data())
def test_safe_min(data: st.DataObject) -> None:
    int_vals = data.draw(st.lists(st.integers()))
    none_vals = data.draw(st.lists(st.none()))
    vals = data.draw(st.permutations(int_vals + none_vals))  # type: ignore

    default_val = data.draw(st.one_of(st.integers(), st.none()))

    result = safe_min(vals, default=default_val)

    if int_vals:
        assert result == min(int_vals)
    else:
        assert result == default_val


@given(st.data())
def test_safe_max(data: st.DataObject) -> None:
    int_vals = data.draw(st.lists(st.integers()))
    none_vals = data.draw(st.lists(st.none()))
    vals = data.draw(st.permutations(int_vals + none_vals))  # type: ignore

    default_val = data.draw(st.one_of(st.integers(), st.none()))

    result = safe_max(vals, default=default_val)

    if int_vals:
        assert result == max(int_vals)
    else:
        assert result == default_val



__all__ = [
    'st_model_prompt',
    'st_model_prompt_list',
    'st_model_run',
    'st_model_run_list',
    'st_model_stdout',
    'st_model_stdout_list',
    'st_model_trace',
    'st_model_trace_list',
    'st_thread_task_no',
    'st_started_at_ended_at',
]

from .st_prompt import st_model_prompt, st_model_prompt_list
from .st_run import st_model_run, st_model_run_list
from .st_stdout import st_model_stdout, st_model_stdout_list
from .st_trace import st_model_trace, st_model_trace_list, st_thread_task_no
from .utils import st_started_at_ended_at



from hypothesis import given, settings
from hypothesis import strategies as st
from sqlalchemy import select
from sqlalchemy.orm import selectinload

from nextline_rdb.db import DB
from nextline_rdb.utils import safe_compare as sc
from nextline_rdb.utils.strategies import (
    st_datetimes,
    st_graphql_ints,
    st_none_or,
    st_ranges,
)

from ... import Model, Run
from .. import st_model_run


@given(st.data())
@settings(max_examples=200)
async def test_options(data: st.DataObject) -> None:
    run_no = data.draw(st_none_or(st_graphql_ints(min_value=1)))
    if run_no is None:
        min_run_no, max_run_no = data.draw(st_ranges(st_=st_graphql_ints, min_start=1))
    else:
        min_run_no, max_run_no = None, None

    min_started_at, max_started_at = data.draw(st_ranges(st_=st_datetimes))
    min_ended_at, max_ended_at = data.draw(
        st_ranges(st_=st_datetimes, min_start=min_started_at)
    )

    generate_traces = data.draw(st.booleans())

    run = data.draw(
        st_model_run(
            run_no=run_no,
            min_run_no=min_run_no,
            max_run_no=max_run_no,
            min_started_at=min_started_at,
            max_started_at=max_started_at,
            min_ended_at=min_ended_at,
            max_ended_at=max_ended_at,
            generate_traces=generate_traces,
        )
    )

    if run_no is not None:
        assert run.run_no == run_no
    else:
        assert sc(min_run_no) <= run.run_no <= sc(max_run_no)

    assert sc(min_started_at) <= sc(run.started_at) <= sc(max_started_at)
    assert sc(min_ended_at) <= sc(run.ended_at) <= sc(max_ended_at)
    assert sc(run.started_at) <= sc(run.ended_at)

    if not run.started_at:
        assert not run.ended_at

    traces = run.traces
    prompts = run.prompts
    stdouts = run.stdouts
    if generate_traces:
        assert traces
    else:
        assert not traces
        assert not prompts
        assert not stdouts

    if run.script:
        compile(run.script, '<string>', 'exec')


@given(run=st_model_run())
async def test_db(run: Run) -> None:
    traces = run.traces
    prompts = run.prompts
    stdouts = run.stdouts

    async with DB(use_migration=False, model_base_class=Model) as db:
        async with db.session.begin() as session:
            session.add(run)
        async with db.session() as session:
            select_run = select(Run)
            run_ = await session.scalar(
                select_run.options(
                    selectinload(Run.traces),
                    selectinload(Run.prompts),
                    selectinload(Run.stdouts),
                )
            )
            assert run_
            session.expunge_all()

    assert repr(run) == repr(run_)
    traces = sorted(traces, key=lambda m: m.id)
    assert repr(traces) == repr(sorted(run_.traces, key=lambda m: m.id))
    prompts = sorted(prompts, key=lambda m: m.id)
    assert repr(prompts) == repr(sorted(run_.prompts, key=lambda m: m.id))
    stdouts = sorted(stdouts, key=lambda m: m.id)
    assert repr(stdouts) == repr(sorted(run_.stdouts, key=lambda m: m.id))



from functools import cmp_to_key, partial

from hypothesis import given, note
from hypothesis import strategies as st

from nextline_rdb.db import DB
from nextline_rdb.pagination import load_models
from nextline_rdb.utils.strategies import st_none_or

from .funcs import cmp, st_entity, st_idx, st_length, st_sort
from .models import Base, Entity


@given(st.data())
async def test_all(data: st.DataObject):
    n_max = 10

    entities = data.draw(st.lists(st_entity(), min_size=0, max_size=n_max))
    sort = data.draw(st_none_or(st_sort()), label='sort')

    async with DB(model_base_class=Base, use_migration=False) as db:
        async with db.session.begin() as session:
            session.add_all(entities)

        note(f'entities={entities}')

        ordered = sorted(entities, key=cmp_to_key(partial(cmp, sort=sort)))

        expected = ordered[:]

        async with db.session() as session:
            models = (await load_models(session, Entity, 'id', sort=sort)).all()

    assert repr(models) == repr(expected)


@given(st.data())
async def test_forward(data: st.DataObject):
    n_max = 10

    entities = data.draw(st.lists(st_entity(), min_size=0, max_size=n_max))
    sort = data.draw(st_none_or(st_sort()), label='sort')

    async with DB(model_base_class=Base, use_migration=False) as db:
        async with db.session.begin() as session:
            session.add_all(entities)

        note(f'entities={entities}')

        ordered = sorted(entities, key=cmp_to_key(partial(cmp, sort=sort)))

        after_idx = data.draw(st_idx(ordered), label='after')
        after = ordered[after_idx].id if after_idx is not None else None
        first = data.draw(
            st_length(len(ordered), allow_none=after_idx is not None), label='first'
        )
        note(f'after={after}, first={first}')

        start = after_idx + 1 if after_idx is not None else None
        end = first if start is None else (None if first is None else (start + first))
        note(f'start={start}, end={end}')

        expected = ordered[slice(start, end, None)]

        async with db.session() as session:
            models = (
                await load_models(
                    session, Entity, 'id', sort=sort, after=after, first=first
                )
            ).all()

    assert repr(models) == repr(expected)


@given(st.data())
async def test_backward(data: st.DataObject):
    n_max = 10

    entities = data.draw(st.lists(st_entity(), min_size=0, max_size=n_max))
    sort = data.draw(st_none_or(st_sort()), label='sort')

    async with DB(model_base_class=Base, use_migration=False) as db:
        async with db.session.begin() as session:
            session.add_all(entities)

        note(f'entities={entities}')

        ordered = sorted(entities, key=cmp_to_key(partial(cmp, sort=sort)))

        before_idx = data.draw(st_idx(ordered), label='before')
        before = ordered[before_idx].id if before_idx is not None else None
        last = data.draw(
            st_length(len(ordered), allow_none=before_idx is not None), label='last'
        )
        note(f'before={before}, last={last}')

        end = before_idx if before_idx is not None else None
        start = (
            None
            if last is None
            else (len(ordered) - last if end is None else max(end - last, 0))
        )
        note(f'start={start}, end={end}')

        expected = ordered[slice(start, end, None)]

        async with db.session() as session:
            models = (
                await load_models(
                    session, Entity, 'id', sort=sort, before=before, last=last
                )
            ).all()

    assert repr(models) == repr(expected)



from datetime import datetime
from typing import TYPE_CHECKING

from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base import Model

if TYPE_CHECKING:
    from .model_run import Run
    from .model_trace import Trace


class Stdout(Model):
    __tablename__ = "stdout"
    id: Mapped[int] = mapped_column(primary_key=True, index=True)
    run_no: Mapped[int]
    trace_no: Mapped[int]
    text: Mapped[str | None]
    written_at: Mapped[datetime | None]

    run_id: Mapped[int] = mapped_column(ForeignKey('run.id'))
    run: Mapped['Run'] = relationship(back_populates='stdouts')

    trace_id: Mapped[int] = mapped_column(ForeignKey('trace.id'))
    trace: Mapped['Trace'] = relationship(back_populates='stdouts')



from hypothesis import given
from hypothesis import strategies as st
from sqlalchemy import select

from nextline_rdb.db import DB
from nextline_rdb.utils.strategies import st_none_or

from ... import Model, Stdout
from .. import st_model_run, st_model_stdout_list, st_model_trace_list


@given(st.data())
async def test_st_model_stdout_lists(data: st.DataObject) -> None:
    # generate_traces=False because that would generate stdout with traces
    if run := data.draw(st_none_or(st_model_run(generate_traces=False))):
        data.draw(st_none_or(st_model_trace_list(run=run, min_size=0, max_size=3)))
        assert run.traces is not None

    max_size = data.draw(st.integers(min_value=0, max_value=5))

    stdouts = data.draw(st_model_stdout_list(run=run, max_size=max_size))

    assert len(stdouts) <= max_size

    if stdouts:
        runs = set(stdout.trace.run for stdout in stdouts)
        assert len(runs) == 1
        assert run is None or run is runs.pop()

    async with DB(use_migration=False, model_base_class=Model) as db:
        async with db.session.begin() as session:
            session.add_all(stdouts)

        async with db.session() as session:
            select_stdout = select(Stdout)
            stdouts_ = (await session.scalars(select_stdout)).all()
            session.expunge_all()
    stdouts = sorted(stdouts, key=lambda stdout: stdout.id)
    assert repr(stdouts) == repr(sorted(stdouts_, key=lambda stdout: stdout.id))



from asyncio import to_thread

from alembic import command
from hypothesis import given, note
from sqlalchemy import select

from nextline_rdb.alembic.models import rev_5a08750d6760 as models_old
from nextline_rdb.alembic.models import rev_f9a742bb2297 as models_new
from nextline_rdb.alembic.models.rev_5a08750d6760.strategies import st_model_run_list
from nextline_rdb.db import DB

from .conftest import AlembicConfigFactory

REVISION_OLD = '5a08750d6760'
REVISION_NEW = 'f9a742bb2297'


@given(runs=st_model_run_list(generate_traces=True))
async def test_migration(
    alembic_config_factory: AlembicConfigFactory, runs: list[models_old.Run]
) -> None:
    note(f'runs: {runs}')

    config = alembic_config_factory()
    url = config.get_main_option('sqlalchemy.url')
    note(f'url: {url}')

    await to_thread(command.upgrade, config, REVISION_OLD)
    async with DB(url, model_base_class=models_old.Model, use_migration=False) as db:
        assert db.migration_revision == REVISION_OLD
        async with db.session.begin() as session:
            session.add_all(runs)

    await to_thread(command.upgrade, config, REVISION_NEW)
    async with DB(url, model_base_class=models_new.Model, use_migration=False) as db:
        assert db.migration_revision == REVISION_NEW

    await to_thread(command.downgrade, config, REVISION_OLD)
    async with DB(url, model_base_class=models_old.Model, use_migration=False) as db:
        assert db.migration_revision == REVISION_OLD
        async with db.session() as session:
            select_run = select(models_old.Run).order_by(models_old.Run.run_no)
            runs_ = (await session.scalars(select_run)).all()

    assert repr(runs) == repr(runs_)



from typing import Callable

import pytest
from alembic.config import Config

from nextline_rdb.db import ALEMBIC_INI

AlembicConfigFactory = Callable[[], Config]


@pytest.fixture(scope='session')
def alembic_config_factory(
    tmp_path_factory: pytest.TempPathFactory,
) -> Callable[[], Config]:
    def factory() -> Config:
        config = Config(ALEMBIC_INI)
        dir = tmp_path_factory.mktemp('db')
        url = f'sqlite+aiosqlite:///{dir}/db.sqlite'
        config.set_main_option('sqlalchemy.url', url)
        return config

    return factory



import datetime
from enum import Enum
from string import ascii_lowercase, ascii_uppercase

from hypothesis import given
from hypothesis import strategies as st

from nextline_rdb.utils.utc import is_timezone_aware

from .. import repr_val


@given(...)
def test_integers(value: int) -> None:
    repr_ = repr_val(value)
    assert eval(repr_) == value


@given(...)
def test_floats(value: float) -> None:
    repr_ = repr_val(value)
    eval_ = eval(repr_)
    if value != value:  # NaN
        assert eval_ != eval_
    else:
        assert eval_ == value


@given(...)
def test_text(value: str) -> None:
    repr_ = repr_val(value)
    assert eval(repr_) == value


class Color(Enum):
    RED = 1
    GREEN = 2
    BLUE = 3


@given(st.sampled_from(Color))
def test_enum(value: Color) -> None:
    repr_ = repr_val(value)
    assert eval(repr_) == value


@st.composite
def st_enum_type(draw: st.DrawFn):
    '''Generate an Enum type.

    >>> enum_type = st_enum_type().example()
    >>> list(enum_type)
    [...]

    Code based on:
    https://github.com/HypothesisWorks/hypothesis/issues/2693#issuecomment-823710924

    '''
    names_ = st.text(ascii_lowercase, min_size=1)
    names = st.builds(lambda x: x.capitalize(), names_).filter(str.isidentifier)
    values = st.lists(st.text(ascii_uppercase, min_size=1), min_size=1, unique=True)
    return draw(st.builds(Enum, names, values))


@given(st.data())
def test_arbitrary_enum(data: st.DataObject):
    enum_type = data.draw(st_enum_type())
    item = data.draw(st.sampled_from(enum_type))
    repr_ = repr_val(item)
    locals()[enum_type.__name__] = enum_type  # so eval can find the type
    eval_ = eval(repr_)
    assert eval_ == item


@given(st.datetimes())
def test_datetimes(value: datetime.datetime) -> None:
    assert not is_timezone_aware(value)
    repr_ = repr_val(value)
    assert datetime  # need for eval
    assert eval(repr_) == value



__all__ = ['register']

from nextline import Nextline

from nextline_rdb.db import DB

from .write_prompt_table import WritePromptTable
from .write_run_table import WriteRunTable
from .write_stdout_table import WriteStdoutTable
from .write_trace_table import WriteTraceTable


def register(nextline: Nextline, db: DB) -> None:
    nextline.register(WriteRunTable(db=db))
    nextline.register(WriteTraceTable(db=db))
    nextline.register(WritePromptTable(db=db))
    nextline.register(WriteStdoutTable(db=db))



"""migrate data

Revision ID: cafceacada62
Revises: 6e3cf7d9b6bf
Create Date: 2024-01-29 16:33:19.473453

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.orm.session import Session

from nextline_rdb.alembic.models.rev_6e3cf7d9b6bf import Run, Script
from nextline_rdb.utils import mark_last

# revision identifiers, used by Alembic.
revision = 'cafceacada62'
down_revision = '6e3cf7d9b6bf'
branch_labels = None
depends_on = None


def upgrade():
    with Session(bind=op.get_bind()) as session:
        with session.begin():
            session.query(Script).delete()

        with session.begin():
            select_runs = sa.select(Run).order_by(Run.run_no)
            runs = session.execute(select_runs).scalars().all()
            scripts = list[Script]()
            for last, run in mark_last(runs):
                if run.script_old is None:
                    continue
                script = Script(script=run.script_old, current=last)
                scripts += [script]
                run.script = script
                run.script_old = None


def downgrade():
    with Session(bind=op.get_bind()) as session:
        with session.begin():
            select_scripts = sa.select(Script).order_by(Script.id)
            scripts = session.execute(select_scripts).scalars().all()
            for script in scripts:
                for run in script.runs:
                    run.script_old = script.script
                session.delete(script)



from datetime import datetime
from typing import TYPE_CHECKING, Optional

from sqlalchemy import ForeignKey, Text
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base import Model

if TYPE_CHECKING:
    from .model_prompt import Prompt
    from .model_script import Script
    from .model_stdout import Stdout
    from .model_trace import Trace


class Run(Model):
    __tablename__ = "run"
    id: Mapped[int] = mapped_column(primary_key=True, index=True)
    run_no: Mapped[int] = mapped_column(unique=True)
    state: Mapped[str | None]
    started_at: Mapped[datetime | None]
    ended_at: Mapped[datetime | None]
    exception: Mapped[str | None] = mapped_column(Text)

    script_id: Mapped[Optional[int]] = mapped_column(
        ForeignKey('script.id'), nullable=True
    )
    script: Mapped[Optional['Script']] = relationship(back_populates='runs')

    traces: Mapped[list["Trace"]] = relationship(
        back_populates='run', cascade='all, delete-orphan'
    )
    prompts: Mapped[list["Prompt"]] = relationship(
        back_populates='run', cascade='all, delete-orphan'
    )
    stdouts: Mapped[list["Stdout"]] = relationship(
        back_populates='run', cascade='all, delete-orphan'
    )



"""init

Revision ID: de90b787e92f
Revises:
Create Date: 2022-03-14 14:00:25.218755

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = 'de90b787e92f'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        'hello',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('message', sa.String(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
    )
    op.create_index(op.f('ix_hello_id'), 'hello', ['id'], unique=False)
    op.create_table(
        'run',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('run_no', sa.Integer(), nullable=False),
        sa.Column('state', sa.String(), nullable=True),
        sa.Column('started_at', sa.DateTime(), nullable=True),
        sa.Column('ended_at', sa.DateTime(), nullable=True),
        sa.Column('script', sa.Text(), nullable=True),
        sa.Column('exception', sa.Text(), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('run_no'),
    )
    op.create_index(op.f('ix_run_id'), 'run', ['id'], unique=False)
    op.create_table(
        'stdout',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('run_no', sa.Integer(), nullable=False),
        sa.Column('text', sa.String(), nullable=True),
        sa.Column('written_at', sa.DateTime(), nullable=True),
        sa.Column('run_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ['run_id'],
            ['run.id'],
        ),
        sa.PrimaryKeyConstraint('id'),
    )
    op.create_index(op.f('ix_stdout_id'), 'stdout', ['id'], unique=False)
    op.create_table(
        'trace',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('run_no', sa.Integer(), nullable=False),
        sa.Column('trace_no', sa.Integer(), nullable=False),
        sa.Column('state', sa.String(), nullable=False),
        sa.Column('thread_no', sa.Integer(), nullable=False),
        sa.Column('task_no', sa.Integer(), nullable=True),
        sa.Column('started_at', sa.DateTime(), nullable=False),
        sa.Column('ended_at', sa.DateTime(), nullable=True),
        sa.Column('run_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ['run_id'],
            ['run.id'],
        ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('run_no', 'trace_no'),
    )
    op.create_index(op.f('ix_trace_id'), 'trace', ['id'], unique=False)
    op.create_table(
        'prompt',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('run_no', sa.Integer(), nullable=False),
        sa.Column('trace_no', sa.Integer(), nullable=False),
        sa.Column('prompt_no', sa.Integer(), nullable=False),
        sa.Column('open', sa.Boolean(), nullable=False),
        sa.Column('event', sa.String(), nullable=False),
        sa.Column('started_at', sa.DateTime(), nullable=False),
        sa.Column('file_name', sa.String(), nullable=True),
        sa.Column('line_no', sa.Integer(), nullable=True),
        sa.Column('stdout', sa.String(), nullable=True),
        sa.Column('command', sa.String(), nullable=True),
        sa.Column('ended_at', sa.DateTime(), nullable=True),
        sa.Column('run_id', sa.Integer(), nullable=False),
        sa.Column('trace_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ['run_id'],
            ['run.id'],
        ),
        sa.ForeignKeyConstraint(
            ['trace_id'],
            ['trace.id'],
        ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('run_no', 'prompt_no'),
    )
    op.create_index(op.f('ix_prompt_id'), 'prompt', ['id'], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f('ix_prompt_id'), table_name='prompt')
    op.drop_table('prompt')
    op.drop_index(op.f('ix_trace_id'), table_name='trace')
    op.drop_table('trace')
    op.drop_index(op.f('ix_stdout_id'), table_name='stdout')
    op.drop_table('stdout')
    op.drop_index(op.f('ix_run_id'), table_name='run')
    op.drop_table('run')
    op.drop_index(op.f('ix_hello_id'), table_name='hello')
    op.drop_table('hello')
    # ### end Alembic commands ###



import strawberry
from hypothesis import given, note
from hypothesis import strategies as st

from nextline_rdb.db import DB
from nextline_rdb.models import Run
from nextline_rdb.models.strategies import st_model_run_list
from nextline_rdb.schema import Query
from nextline_rdb.utils.strategies import st_graphql_ints, st_none_or

from ..graphql import QUERY_RDB_RUN


@st.composite
def st_query_variables(
    draw: st.DrawFn, runs: list[Run]
) -> tuple[dict[str, int | None], Run | None]:
    run = draw(st_none_or(st.sampled_from(runs)) if runs else st.none())
    if run:
        id = draw(st.sampled_from([None, run.id]))
        run_no = run.run_no if id is None else None
        variables = {'id': id, 'runNo': run_no}
        return variables, run

    id = draw(
        st_none_or(
            st_graphql_ints(min_value=1).filter(
                lambda i: i not in {run.id for run in runs}
            )
        )
    )
    run_no = (
        None
        if id is not None
        else draw(
            st_graphql_ints(min_value=1).filter(
                lambda i: i not in {run.run_no for run in runs}
            )
        )
    )
    variables = {'id': id, 'runNo': run_no}
    return variables, None


@given(st.data())
async def test_run(data: st.DataObject) -> None:
    schema = strawberry.Schema(query=Query)

    runs = data.draw(st_model_run_list(generate_traces=False, min_size=0, max_size=5))

    async with DB() as db:
        async with db.session.begin() as session:
            session.add_all(runs)
        note(f'runs: {runs}')

        variables, run = data.draw(st_query_variables(runs=runs))

        resp = await schema.execute(
            QUERY_RDB_RUN, variable_values=variables, context_value={'db': db}
        )
        note(f'resp: {resp}')

        assert resp.data

        run_ = resp.data['rdb']['run']

        if run is None:
            assert run_ is None
        else:
            assert run_
            assert run_['id'] == run.id
            assert run_['runNo'] == run.run_no



'''Pagination based on the example code in strawberry doc

Strawberry doc: https://strawberry.rocks/docs/guides/pagination/connections
Relay doc: https://relay.dev/graphql/connections.htm
'''

from collections.abc import Callable, Coroutine
from typing import Any, Generic, Optional, TypeVar

import strawberry

_T = TypeVar('_T')


@strawberry.type
class Edge(Generic[_T]):
    node: _T
    cursor: str


@strawberry.type
class PageInfo:
    has_next_page: bool
    has_previous_page: bool
    start_cursor: Optional[str] = None
    end_cursor: Optional[str] = None


@strawberry.type
class Connection(Generic[_T]):
    page_info: PageInfo
    edges: list[Edge[_T]]


async def query_connection(
    query_edges: Callable[..., Coroutine[Any, Any, list[Edge[_T]]]],
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> Connection[_T]:
    forward = after or (first is not None)
    backward = before or (last is not None)

    if forward and backward:
        raise ValueError('Only either after/first or before/last is allowed')

    if forward:
        if first is not None:
            first += 1  # add one for has_next_page
        edges = await query_edges(after=after, first=first)
        has_previous_page = not not after
        if has_next_page := len(edges) == first:
            edges = edges[:-1]
    elif backward:
        if last is not None:
            last += 1  # add one for has_previous_page
        edges = await query_edges(before=before, last=last)
        if has_previous_page := len(edges) == last:
            edges = edges[1:]
        has_next_page = not not before
    else:
        edges = await query_edges()
        has_previous_page = False
        has_next_page = False

    page_info = PageInfo(
        has_previous_page=has_previous_page,
        has_next_page=has_next_page,
        start_cursor=edges[0].cursor if edges else None,
        end_cursor=edges[-1].cursor if edges else None,
    )

    return Connection(page_info=page_info, edges=edges)



from collections.abc import Callable

import pytest
from hypothesis import given
from hypothesis import strategies as st

from nextline_rdb.db import DB
from nextline_rdb.utils import class_name_and_primary_keys_of, ensure_sync_url, load_all

from .models import Bar, Foo, Model, register_session_events


async def test_ensure_sync_url(tmp_url_factory: Callable[[], str]):
    url = tmp_url_factory()
    sync_url = ensure_sync_url(url)

    async with DB(url=sync_url) as db:
        assert db.url == url


async def test_fields():
    db = DB()
    assert db.url
    assert db.metadata
    assert db.engine

    async with db:
        assert db.session


@given(st.booleans())
async def test_migration_revision(use_migration: bool) -> None:
    async with DB(use_migration=use_migration) as db:
        if use_migration:
            assert db.migration_revision
        else:
            assert db.migration_revision is None


@given(st.lists(st.integers(min_value=0, max_value=4), min_size=0, max_size=4))
async def test_session_nested(tmp_url_factory: Callable[[], str], sizes: list[int]):
    url = tmp_url_factory()

    objs = [Foo(bars=[Bar() for _ in range(size)]) for size in sizes]

    async with DB(
        url=url,
        model_base_class=Model,
        use_migration=False,
        register_session_events=register_session_events,
    ) as db:
        async with db.session() as session:
            async with session.begin():
                session.add_all(objs)  # Related objects will be automatically added.
                added = list(session.new)
                assert set(objs) <= set(added)

            added = sorted(added, key=class_name_and_primary_keys_of)
            assert len(added) == sum(sizes) + len(sizes)

            repr_added = [repr(m) for m in added]

        async with db.session() as session:
            loaded = await load_all(session, Model)
            repr_loaded = [repr(m) for m in loaded]

        assert repr_added == repr_loaded


@given(st.lists(st.integers(min_value=0, max_value=4), min_size=0, max_size=4))
async def test_session_begin(tmp_url_factory: Callable[[], str], sizes: list[int]):
    url = tmp_url_factory()

    objs = [Foo(bars=[Bar() for _ in range(size)]) for size in sizes]

    async with DB(
        url=url,
        model_base_class=Model,
        use_migration=False,
        register_session_events=register_session_events,
    ) as db:
        async with db.session.begin() as session:
            session.add_all(objs)  # Related objects will be automatically added.
            added = list(session.new)
            assert set(objs) <= set(added)

        added = sorted(added, key=class_name_and_primary_keys_of)
        assert len(added) == sum(sizes) + len(sizes)

        repr_added = [repr(m) for m in added]

        async with db.session() as session:
            loaded = await load_all(session, Model)
            repr_loaded = [repr(m) for m in loaded]

        assert repr_added == repr_loaded


@pytest.fixture(scope='session')
def tmp_url_factory(tmp_path_factory: pytest.TempPathFactory) -> Callable[[], str]:
    def factory() -> str:
        dir = tmp_path_factory.mktemp('db')
        url = f'sqlite+aiosqlite:///{dir}/db.sqlite'
        return url

    return factory



from asyncio import to_thread

from alembic import command
from hypothesis import given, note
from hypothesis import strategies as st

from nextline_rdb.alembic.models import rev_f9a742bb2297 as models_old
from nextline_rdb.alembic.models.rev_f9a742bb2297.strategies import st_model_script
from nextline_rdb.db import DB
from nextline_rdb.utils import load_all

from .conftest import AlembicConfigFactory

REVISION_START = 'f9a742bb2297'
REVISION_STAGE_ONE = 'c72fa3ee6a1a'
REVISION_NEW = '4dc6a93dfed8'


@given(scripts=st.lists(st_model_script(), min_size=0, max_size=5))
async def test_migration(
    alembic_config_factory: AlembicConfigFactory, scripts: list[models_old.Script]
) -> None:
    note(f'scripts: {scripts}')

    config = alembic_config_factory()
    url = config.get_main_option('sqlalchemy.url')
    note(f'url: {url}')

    # Enter data, in which multiple scripts can be current
    await to_thread(command.upgrade, config, REVISION_START)
    async with DB(url, model_base_class=models_old.Model, use_migration=False) as db:
        assert db.migration_revision == REVISION_START
        async with db.session.begin() as session:
            session.add_all(scripts)

    # Upgrade to the revision where at most one script is current. Make expected data.
    await to_thread(command.upgrade, config, REVISION_STAGE_ONE)
    async with DB(url, model_base_class=models_old.Model, use_migration=False) as db:
        assert db.migration_revision == REVISION_STAGE_ONE
        async with db.session() as session:
            loaded = await load_all(session, models_old.Model)
            expected = [repr(s) for s in loaded]

    # Upgrade to the target revision.
    await to_thread(command.upgrade, config, REVISION_NEW)

    # Downgrade and make actual data.
    await to_thread(command.downgrade, config, REVISION_START)
    async with DB(url, model_base_class=models_old.Model, use_migration=False) as db:
        assert db.migration_revision == REVISION_START
        async with db.session() as session:
            loaded = await load_all(session, models_old.Model)
            actual = [repr(s) for s in loaded]

    assert expected == actual



from typing import Optional

from hypothesis import strategies as st

from .. import Model


@st.composite
def st_model_instance_list(
    draw: st.DrawFn,
    min_size: int = 0,
    max_size: Optional[int] = None,
) -> list[Model]:
    '''A strategy for a list of `Model` instances.

    Instances of `Script` and `Run` models are explicitly generated.  Instances of other
    models will be generated as related models of `Script` and `Run` models.
    '''
    from .st_run import st_model_run_list
    from .st_script import st_model_script_list

    # Generate a list of `Script` models
    scripts = draw(st_model_script_list(min_size=min_size, max_size=max_size))

    # Adjust min_size and max_size
    size = len(scripts)
    min_size = max(0, min_size - size)
    max_size = max_size - size if max_size is not None else None

    # Generate a list of `Run` models
    runs = draw(
        st_model_run_list(
            generate_traces=True,
            min_size=min_size,
            max_size=max_size,
            scripts=scripts,
        )
    )

    ret: list[Model] = list(scripts) + list(runs)
    return ret



from typing import Optional, cast

import strawberry
from sqlalchemy import select
from sqlalchemy.orm import selectinload
from strawberry.types import Info

import nextline_rdb
from nextline_rdb.db import DB
from nextline_rdb.models import Run

from . import types
from .pagination import Connection


async def query_run(
    info: Info, id: Optional[int] = None, run_no: Optional[int] = None
) -> types.RunHistory | None:
    db = cast(DB, info.context['db'])
    async with db.session() as session:
        stmt = select(Run).options(selectinload(Run.script))
        if id is not None:
            stmt = stmt.filter(Run.id == id)
        else:
            stmt = stmt.filter(Run.run_no == run_no)
        run = (await session.execute(stmt)).scalar_one_or_none()
    return types.RunHistory.from_model(run) if run else None


def query_migration_version(info: Info) -> str | None:
    db = cast(DB, info.context['db'])
    return db.migration_revision


@strawberry.type
class QueryRDB:
    runs: Connection[types.RunHistory] = strawberry.field(
        resolver=types.query_connection_run
    )
    traces: Connection[types.TraceHistory] = strawberry.field(
        resolver=types.query_connection_trace
    )
    prompts: Connection[types.PromptHistory] = strawberry.field(
        resolver=types.query_connection_prompt
    )
    stdouts: Connection[types.StdoutHistory] = strawberry.field(
        resolver=types.query_connection_stdout
    )
    run: types.RunHistory | None = strawberry.field(resolver=query_run)
    version: str = nextline_rdb.__version__
    migration_version: str | None = strawberry.field(resolver=query_migration_version)


@strawberry.type
class Query:
    @strawberry.field
    async def history(self) -> QueryRDB:
        # TODO: Remove this method.
        return QueryRDB()

    @strawberry.field
    async def rdb(self) -> QueryRDB:
        return QueryRDB()



"""add the table "script"

Revision ID: 6e3cf7d9b6bf
Revises: 2fa41efc56f0
Create Date: 2024-01-29 15:11:23.479015

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = '6e3cf7d9b6bf'
down_revision = '2fa41efc56f0'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute('PRAGMA foreign_keys=OFF;')
    op.create_table('script',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('current', sa.Boolean(), nullable=False),
    sa.Column('script', sa.Text(), nullable=False),
    sa.PrimaryKeyConstraint('id', name=op.f('pk_script'))
    )
    with op.batch_alter_table('script', schema=None) as batch_op:
        batch_op.create_index(batch_op.f('ix_script_id'), ['id'], unique=False)

    with op.batch_alter_table('run', schema=None) as batch_op:
        batch_op.add_column(sa.Column('script_id', sa.Integer(), nullable=True))
        batch_op.create_foreign_key(batch_op.f('fk_run_script_id_script'), 'script', ['script_id'], ['id'])
    op.execute('PRAGMA foreign_keys=ON;')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute('PRAGMA foreign_keys=OFF;')
    with op.batch_alter_table('run', schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f('fk_run_script_id_script'), type_='foreignkey')
        batch_op.drop_column('script_id')

    with op.batch_alter_table('script', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_script_id'))

    op.drop_table('script')
    op.execute('PRAGMA foreign_keys=ON;')
    # ### end Alembic commands ###



import enum
import math
from collections.abc import Iterable
from typing import Any

from sqlalchemy.orm import ColumnProperty, class_mapper


class ReprMixin:
    '''A mixin for ORM models with __repr__ for object reconstruction.

    Override methods of this class if necessary.
    '''

    def __repr__(self) -> str:
        '''Return a string that can be used to reconstruct the object.'''
        class_name = self._repr_class_name()
        init_args = self._repr_init_args()
        return f'{class_name}({init_args})'

    def _repr_class_name(self) -> str:
        '''The class name to use in the repr.'''
        return self.__class__.__qualname__

    def _repr_init_args(self) -> str:
        '''The arguments to use in the repr.'''
        keys = self._repr_keys()
        fmts = (f'{k}={self._repr_val(k)}' for k in keys)
        return ', '.join(fmts)

    def _repr_keys(self) -> Iterable[str]:
        '''The attribute names to include in the repr.'''
        props = class_mapper(self.__class__).iterate_properties
        columns = (p for p in props if isinstance(p, ColumnProperty))
        return (c.key for c in columns if self._repr_filter(c.key))

    def _repr_filter(self, key: str) -> bool:
        '''Exclude if the value is None.'''
        val = getattr(self, key)
        return val is not None

    def _repr_val(self, key: str) -> str:
        '''Format the value so it can be reconstructed.'''
        val = getattr(self, key)
        return repr_val(val)


def repr_val(val: Any) -> str:
    match val:
        case enum.Enum():
            return str(val)
        case int():
            return f'{val:_}'
        case math.inf:
            return 'float("inf")'
        case float(neg_inf) if neg_inf == float('-inf'):
            return 'float("-inf")'
        case float(nan_val) if math.isnan(nan_val):
            return 'float("nan")'
        case _:
            return repr(val)



from sqlalchemy import MetaData
from sqlalchemy.orm import DeclarativeBase

from .convention import NAMING_CONVENTION
from .repr_ import ReprMixin

metadata = MetaData(naming_convention=dict(NAMING_CONVENTION))


class Model(ReprMixin, DeclarativeBase):
    metadata = metadata



from typing import NamedTuple, Optional, Type, TypeVar, cast

from sqlalchemy import func, select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import DeclarativeBase, aliased, selectinload
from sqlalchemy.sql.expression import literal
from sqlalchemy.sql.selectable import Select

# import sqlparse


# def format_sql(sql):
#     return sqlparse.format(str(sql), reindent=True, keyword_case='upper')


class SortField(NamedTuple):
    field: str
    desc: bool = False


Sort = list[SortField]

_Id = TypeVar('_Id')


async def load_models(
    session: AsyncSession,
    Model: Type[DeclarativeBase],
    id_field: str,
    *,
    sort: Optional[Sort] = None,
    before: Optional[_Id] = None,
    after: Optional[_Id] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
):
    stmt = compose_statement(
        Model,
        id_field,
        sort=sort,
        before=before,
        after=after,
        first=first,
        last=last,
    )

    # TODO: Make this optional
    stmt = stmt.options(selectinload('*'))

    models = await session.scalars(stmt)
    return models


def compose_statement(
    Model: Type[DeclarativeBase],
    id_field: str,
    *,
    sort: Optional[Sort] = None,
    before: Optional[_Id] = None,
    after: Optional[_Id] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> Select:
    '''Return a SELECT statement object to be given to session.scalars'''

    forward = (after is not None) or (first is not None)
    backward = (before is not None) or (last is not None)

    if forward and backward:
        raise ValueError('Only either after/first or before/last is allowed')

    sort = sort or []

    if id_field not in [s.field for s in sort]:
        sort.append(SortField(id_field))

    def sorting_fields(Model, reverse=False):
        return [
            f.desc() if reverse ^ d else f
            for f, d in [(getattr(Model, s.field), s.desc) for s in sort]
        ]

    if not (forward or backward):
        return select(Model).order_by(*sorting_fields(Model))

    cursor = after if forward else before
    limit = first if forward else last

    if cursor is None:
        stmt = select(Model).order_by(*sorting_fields(Model, reverse=backward))
    else:
        cte = select(
            Model,
            func.row_number()
            .over(order_by=sorting_fields(Model, reverse=backward))
            .label('row_number'),
        ).cte()

        subq = select(cte.c.row_number.label('cursor'))
        subq = subq.where(getattr(cte.c, id_field) == cursor)
        subq = cast(Select[tuple], subq.subquery())

        Alias = aliased(Model, cte)  # type: ignore
        stmt = select(Alias).select_from(cte)
        stmt = stmt.join(subq, literal(True))  # type: ignore # cartesian product
        stmt = stmt.order_by(*sorting_fields(Alias, reverse=backward))
        stmt = stmt.where(cte.c.row_number > subq.c.cursor)

    if limit is not None:
        stmt = stmt.limit(limit)

    if backward:
        Alias = aliased(Model, stmt.subquery())
        stmt = select(Alias).order_by(*sorting_fields(Alias))

    return stmt



from hypothesis import given
from hypothesis import strategies as st
from sqlalchemy import select

from nextline_rdb.db import DB
from nextline_rdb.utils import safe_compare as sc
from nextline_rdb.utils.strategies import st_none_or
from nextline_rdb.utils.strategies.misc import st_graphql_ints

from ... import Model, Trace
from .. import st_model_run, st_model_trace, st_thread_task_no


@given(st.data())
async def test_st_model_trace(data: st.DataObject) -> None:
    run = data.draw(st_none_or(st_model_run(generate_traces=False)))
    trace_no = data.draw(st_none_or(st_graphql_ints(min_value=1)))
    thread_task_no = data.draw(st_none_or(st_thread_task_no()))
    generate_prompts = False if run else data.draw(st.booleans())

    trace = data.draw(
        st_model_trace(
            run=run,
            trace_no=trace_no,
            thread_task_no=thread_task_no,
            generate_prompts=generate_prompts,
        )
    )

    assert run is None or run is trace.run
    run = trace.run

    assert trace_no is None or trace_no == trace.trace_no
    assert thread_task_no is None or thread_task_no == (trace.thread_no, trace.task_no)

    assert sc(run.started_at) <= trace.started_at
    assert trace.started_at <= sc(trace.ended_at)
    assert sc(trace.ended_at) <= sc(run.ended_at)

    assert not generate_prompts or trace.prompts

    async with DB(use_migration=False, model_base_class=Model) as db:
        async with db.session.begin() as session:
            session.add(trace)
        async with db.session() as session:
            stmt = select(Trace)
            trace_ = await session.scalar(stmt)
            session.expunge_all()
    assert repr(trace) == repr(trace_)



import asyncio


def run():
    asyncio.run(c())
    return


async def c():
    a = asyncio.create_task(d())
    await asyncio.sleep(0.001)
    await asyncio.sleep(0)
    await a
    return


async def d():
    await asyncio.sleep(0)
    await asyncio.sleep(0.001)
    return



from typing import Optional

from hypothesis import strategies as st

from nextline_rdb.utils.strategies import st_python_scripts

from .. import Script


def st_model_script(current: Optional[bool] = None) -> st.SearchStrategy[Script]:
    st_current = st.just(current) if current is not None else st.booleans()
    return st.builds(Script, current=st_current, script=st_python_scripts())



from __future__ import annotations

import datetime
from typing import Optional, Type, TypeVar

import strawberry
from sqlalchemy import inspect
from strawberry.types import Info

from nextline_rdb import models as db_models
from nextline_rdb.pagination import Sort, SortField

from .pagination import Connection, load_connection


async def query_connection_run(
    info: Info,
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> Connection[RunHistory]:
    sort = [SortField('run_no')]
    Model = db_models.Run
    NodeType = RunHistory
    return await query_connection(
        info, sort, before, after, first, last, Model, NodeType
    )


async def query_connection_trace(
    info: Info,
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> Connection[TraceHistory]:
    sort = [SortField('run_no'), SortField('trace_no')]
    Model = db_models.Trace
    NodeType = TraceHistory
    return await query_connection(
        info, sort, before, after, first, last, Model, NodeType
    )


async def query_connection_prompt(
    info: Info,
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> Connection[PromptHistory]:
    sort = [SortField('run_no'), SortField('prompt_no')]
    Model = db_models.Prompt
    NodeType = PromptHistory
    return await query_connection(
        info, sort, before, after, first, last, Model, NodeType
    )


async def query_connection_stdout(
    info: Info,
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
) -> Connection[StdoutHistory]:
    sort = [SortField('run_no'), SortField('id')]
    Model = db_models.Stdout
    NodeType = StdoutHistory
    return await query_connection(
        info, sort, before, after, first, last, Model, NodeType
    )


_T = TypeVar("_T")


async def query_connection(
    info: Info,
    sort: Optional[Sort],
    before: Optional[str],
    after: Optional[str],
    first: Optional[int],
    last: Optional[int],
    Model,
    NodeType: Type[_T],
) -> Connection[_T]:
    id_field = inspect(Model).primary_key[0].name
    # TODO: handle multiple primary keys

    db = info.context['db']
    create_node_from_model = NodeType.from_model  # type: ignore

    return await load_connection(
        db,
        Model,
        id_field,
        create_node_from_model,
        sort=sort,
        before=before,
        after=after,
        first=first,
        last=last,
    )


@strawberry.type
class RunHistory:
    _model: strawberry.Private[db_models.Run]
    id: int
    run_no: int
    state: Optional[str]
    started_at: Optional[datetime.datetime]
    ended_at: Optional[datetime.datetime]
    script: Optional[str]
    exception: Optional[str]

    # traces: Connection[TraceHistory] = strawberry.field(
    #     resolver=query_connection_trace
    # )
    # prompts: Connection[PromptHistory] = strawberry.field(
    #     resolver=query_connection_prompt
    # )
    # stdouts: Connection[StdoutHistory] = strawberry.field(
    #     resolver=query_connection_stdout
    # )

    @strawberry.field
    def traces(self) -> list["TraceHistory"]:
        return [TraceHistory.from_model(m) for m in self._model.traces]  # type: ignore

    @strawberry.field
    def prompts(self) -> list["PromptHistory"]:
        return [PromptHistory.from_model(m) for m in self._model.prompts]  # type: ignore

    @strawberry.field
    def stdouts(self) -> list["StdoutHistory"]:
        return [StdoutHistory.from_model(m) for m in self._model.stdouts]  # type: ignore

    @classmethod
    def from_model(cls: Type["RunHistory"], model: db_models.Run):
        script = model.script.script if model.script else None
        return cls(
            _model=model,
            id=model.id,
            run_no=model.run_no,
            state=model.state,
            started_at=model.started_at,
            ended_at=model.ended_at,
            script=script,
            exception=model.exception,
        )


@strawberry.type
class TraceHistory:
    _model: strawberry.Private[db_models.Trace]
    id: int
    run_no: int
    trace_no: int
    state: str
    thread_no: int
    started_at: datetime.datetime
    task_no: Optional[int]
    ended_at: Optional[datetime.datetime]

    @strawberry.field
    def run(self) -> RunHistory:
        return RunHistory.from_model(self._model.run)

    # prompts: Connection[PromptHistory] = strawberry.field(
    #     resolver=query_connection_prompt
    # )

    # stdouts: Connection[StdoutHistory] = strawberry.field(
    #     resolver=query_connection_stdout
    # )

    @strawberry.field
    def prompts(self) -> list["PromptHistory"]:
        return [PromptHistory.from_model(m) for m in self._model.prompts]  # type: ignore

    @strawberry.field
    def stdouts(self) -> list["StdoutHistory"]:
        return [StdoutHistory.from_model(m) for m in self._model.stdouts]  # type: ignore

    @classmethod
    def from_model(cls: Type[TraceHistory], model: db_models.Trace):
        return cls(
            _model=model,
            id=model.id,
            run_no=model.run_no,
            trace_no=model.trace_no,
            state=model.state,
            thread_no=model.thread_no,
            started_at=model.started_at,
            task_no=model.task_no,
            ended_at=model.ended_at,
        )


@strawberry.type
class PromptHistory:
    _model: strawberry.Private[db_models.Prompt]
    id: int
    run_no: int
    trace_no: int
    prompt_no: int
    open: bool
    event: str
    started_at: datetime.datetime
    file_name: Optional[str] = None
    line_no: Optional[int] = None
    stdout: Optional[str] = None
    command: Optional[str] = None
    ended_at: Optional[datetime.datetime] = None

    @strawberry.field
    def run(self) -> RunHistory:
        return RunHistory.from_model(self._model.run)

    @strawberry.field
    def trace(self) -> TraceHistory:
        return TraceHistory.from_model(self._model.trace)

    @classmethod
    def from_model(cls: Type[PromptHistory], model: db_models.Prompt):
        return cls(
            _model=model,
            id=model.id,
            run_no=model.run_no,
            trace_no=model.trace_no,
            prompt_no=model.prompt_no,
            open=model.open,
            event=model.event,
            started_at=model.started_at,
            file_name=model.file_name,
            line_no=model.line_no,
            stdout=model.stdout,
            command=model.command,
            ended_at=model.ended_at,
        )


@strawberry.type
class StdoutHistory:
    _model: strawberry.Private[db_models.Stdout]
    id: int
    run_no: int
    trace_no: int
    text: Optional[str] = None
    written_at: Optional[datetime.datetime] = None

    @strawberry.field
    def run(self) -> RunHistory:
        return RunHistory.from_model(self._model.run)

    @strawberry.field
    def trace(self) -> TraceHistory:
        return TraceHistory.from_model(self._model.trace)

    @classmethod
    def from_model(cls: Type[StdoutHistory], model: db_models.Stdout):
        return cls(
            _model=model,
            id=model.id,
            run_no=model.run_no,
            trace_no=model.trace_no,
            text=model.text,
            written_at=model.written_at,
        )



import strawberry
from hypothesis import given

from nextline_rdb.db import DB
from nextline_rdb.models import Run
from nextline_rdb.models.strategies import st_model_run_list
from nextline_rdb.schema import Query

from ..graphql import QUERY_RDB_CONNECTIONS


@given(runs=st_model_run_list(generate_traces=True, min_size=0, max_size=3))
async def test_history(runs: list[Run]) -> None:
    schema = strawberry.Schema(query=Query)

    async with DB() as db:
        async with db.session.begin() as session:
            session.add_all(runs)

        resp = await schema.execute(QUERY_RDB_CONNECTIONS, context_value={'db': db})

    assert resp.data
    rdb = resp.data['rdb']
    runs_ = rdb['runs']['edges']

    assert len(runs) == len(runs_)

    # TODO: Now only the number of runs is checked. Check the contents as well.



import asyncio

from nextline.events import OnWriteStdout
from nextline.plugin.spec import hookimpl
from sqlalchemy import select

from nextline_rdb.db import DB
from nextline_rdb.models import Run, Stdout, Trace


class WriteStdoutTable:
    def __init__(self, db: DB) -> None:
        self._db = db

    @hookimpl
    async def on_write_stdout(self, event: OnWriteStdout) -> None:
        async with self._db.session.begin() as session:
            select_runs = select(Run).filter_by(run_no=event.run_no)
            while not (
                run := (await session.execute(select_runs)).scalar_one_or_none()
            ):
                await asyncio.sleep(0)
            select_traces = select(Trace).filter_by(
                run_no=event.run_no, trace_no=event.trace_no
            )
            while not (
                trace := (await session.execute(select_traces)).scalar_one_or_none()
            ):
                await asyncio.sleep(0)
            stdout = Stdout(
                run_no=event.run_no,
                trace_no=event.trace_no,
                text=event.text,
                written_at=event.written_at,
                run=run,
                trace=trace,
            )
            session.add(stdout)



import time

time.sleep(0.001)


def f():
    for _ in range(10):
        pass
    return


f()
f()

print("here!")

import script_threading  # noqa: E402

script_threading.run()  # step

import script_asyncio  # noqa: E402

script_asyncio.run()  # step



# Constraints naming conventions
#
# SQLAlchemy 2 Doc:
# https://docs.sqlalchemy.org/en/20/core/constraints.html#configuring-constraint-naming-conventions
#
# About the change on 'ck':
# https://stackoverflow.com/a/56000475/7309855
#
# Equivalent code in ProductDB:
# https://github.com/simonsobs/acondbs/blob/7b4e5ab967ce/acondbs/db/sa.py
NAMING_CONVENTION = {
    'ix': 'ix_%(column_0_label)s',
    'uq': 'uq_%(table_name)s_%(column_0_name)s',
    # 'ck': 'ck_%(table_name)s_%(constraint_name)s',
    'ck': 'ck_%(table_name)s_%(column_0_name)s',
    'fk': 'fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s',
    'pk': 'pk_%(table_name)s',
}



__all__ = [
    'NAMING_CONVENTION',
    'Model',
    'ReprMixin',
    'repr_val',
    'Hello',
    'Run',
    'Trace',
    'Prompt',
    'Script',
    'Stdout',
]


from .base import NAMING_CONVENTION, Model, ReprMixin, repr_val
from .model_hello import Hello
from .model_prompt import Prompt
from .model_run import Run
from .model_script import Script
from .model_stdout import Stdout
from .model_trace import Trace



"""rename the filed script to script_old on the table run

Revision ID: 2fa41efc56f0
Revises: 5a08750d6760
Create Date: 2024-01-29 14:12:56.054544

"""
from alembic import op


# revision identifiers, used by Alembic.
revision = '2fa41efc56f0'
down_revision = '5a08750d6760'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute('PRAGMA foreign_keys=OFF;')
    with op.batch_alter_table('run', schema=None) as batch_op:
        batch_op.alter_column('script', new_column_name='script_old')
    op.execute('PRAGMA foreign_keys=ON;')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute('PRAGMA foreign_keys=OFF;')
    with op.batch_alter_table('run', schema=None) as batch_op:
        batch_op.alter_column('script_old', new_column_name='script')
    op.execute('PRAGMA foreign_keys=ON;')
    # ### end Alembic commands ###



from asyncio import to_thread

from alembic import command
from hypothesis import given, note
from hypothesis import strategies as st
from sqlalchemy import select

from nextline_rdb.alembic.models import rev_f9a742bb2297 as models_old
from nextline_rdb.alembic.models.rev_f9a742bb2297.strategies import st_model_script
from nextline_rdb.db import DB

from .conftest import AlembicConfigFactory

REVISION_OLD = 'f9a742bb2297'
REVISION_NEW = 'c72fa3ee6a1a'


@given(scripts=st.lists(st_model_script(), min_size=0, max_size=5))
async def test_migration(
    alembic_config_factory: AlembicConfigFactory, scripts: list[models_old.Script]
) -> None:
    note(f'scripts: {scripts}')

    config = alembic_config_factory()
    url = config.get_main_option('sqlalchemy.url')
    note(f'url: {url}')

    await to_thread(command.upgrade, config, REVISION_OLD)
    async with DB(url, model_base_class=models_old.Model, use_migration=False) as db:
        assert db.migration_revision == REVISION_OLD
        async with db.session.begin() as session:
            session.add_all(scripts)

    await to_thread(command.upgrade, config, REVISION_NEW)

    await to_thread(command.downgrade, config, REVISION_OLD)
    async with DB(url, model_base_class=models_old.Model, use_migration=False) as db:
        assert db.migration_revision == REVISION_OLD
        async with db.session() as session:
            select_scripts = select(models_old.Script).order_by(models_old.Script.id)
            scripts_ = (await session.execute(select_scripts)).scalars().all()
            currents = [s for s in scripts_ if s.current]
            assert len(currents) in (0, 1)



import datetime as dt
from typing import Optional

from hypothesis import strategies as st

from nextline_rdb.utils.strategies import st_datetimes, st_ranges


def st_started_at_ended_at(
    min_start: Optional[dt.datetime] = None,
    max_start: Optional[dt.datetime] = None,
    min_end: Optional[dt.datetime] = None,
    max_end: Optional[dt.datetime] = None,
    allow_start_none: bool = True,
    allow_end_none: bool = True,
    allow_equal: bool = True,
) -> st.SearchStrategy[tuple[Optional[dt.datetime], Optional[dt.datetime]]]:
    '''Generate two naive datetime objects: started_at and ended_at.

    `started_at` and `ended_at` can be `None`. If `started_at` is `None`,
    `ended_at` is always `None`.

    When neither of them is `None`, `started_at` is earlier than or the same
    time as `ended_at`.

    >>> st_started_at_ended_at().example()
    (...)

    '''
    return st_ranges(
        st_=st_datetimes,
        min_start=min_start,
        max_start=max_start,
        min_end=min_end,
        max_end=max_end,
        allow_start_none=allow_start_none,
        allow_end_none=allow_end_none,
        allow_equal=allow_equal,
        let_end_none_if_start_none=True,
    )



from datetime import datetime
from typing import TYPE_CHECKING

from sqlalchemy import ForeignKey, UniqueConstraint
from sqlalchemy.orm import Mapped, mapped_column, relationship

from .base import Model

if TYPE_CHECKING:
    from .model_prompt import Prompt
    from .model_run import Run
    from .model_stdout import Stdout


class Trace(Model):
    __tablename__ = "trace"
    id: Mapped[int] = mapped_column(primary_key=True, index=True)
    run_no: Mapped[int]
    trace_no: Mapped[int]
    state: Mapped[str]
    thread_no: Mapped[int]
    task_no: Mapped[int | None]
    started_at: Mapped[datetime]
    ended_at: Mapped[datetime | None]

    run_id: Mapped[int] = mapped_column(ForeignKey('run.id'))
    run: Mapped['Run'] = relationship(back_populates='traces')

    prompts: Mapped[list["Prompt"]] = relationship(back_populates="trace")
    stdouts: Mapped[list["Stdout"]] = relationship(back_populates="trace")

    __table_args__ = (UniqueConstraint("run_no", "trace_no"),)
